#!/usr/bin/env python3
"""
File adapter for the requests library.
"""

import datetime
import errno
import io
import logging
import mimetypes
import os
import pathlib

from requests import Response, codes
from requests.adapters import BaseAdapter
from requests.compat import urlparse

from .ip import is_localhost


LOGGER = logging.getLogger(__name__)


class FileAdapter(BaseAdapter):
    """
    File adapter for the requests package.
    """

    ERROR_MAP = {
        # pylint: disable=no-member
        errno.EACCES: codes.forbidden,
        errno.ENOENT: codes.not_found,
    }

    def send(
        self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None
    ):  # pylint: disable=too-many-arguments
        """
        Send a file via a requests.Response object.

        Args:
            request:
                The sent request.

            stream:
                If True, stream the file content.

            timeout:
                Ignored.

            verify:
                Ignored

            cert:
                Ignored.

            proxies: Ignored.
        """
        supported_methods = ("GET", "HEAD")
        if request.method not in supported_methods:
            raise ValueError(
                f"Unsupported request method: {request.method} "
                f"[supported methods: {', '.join(supported_methods)}]"
            )

        parsed_url = urlparse(request.url)

        if parsed_url.netloc and not is_localhost(parsed_url.netloc):
            raise ValueError(
                f'The host "{parsed_url.netloc}" ' "does not resolve to localhost."
            )

        path = pathlib.Path(parsed_url.path)
        if not path.is_absolute():
            raise ValueError("Only absolute paths are supported.")

        LOGGER.debug("Responding to request for file %s", path)

        resp = Response()
        resp.request = request

        try:
            if request.method == "GET":
                resp.raw = path.open("rb")  # pylint: disable=consider-using-with
                resp.raw.release_conn = resp.raw.close
            elif not path.exists():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), str(path)
                )
            elif not os.access(path, os.R_OK):
                raise PermissionError(
                    errno.EACCES, os.strerror(errno.EACCES), str(path)
                )
            resp.status_code = codes.ok  # pylint: disable=no-member
            resp.url = request.url
            if path.is_file():
                self._set_file_headers(resp, path)

        except OSError as err:
            resp.status_code = self.ERROR_MAP.get(
                err.errno, codes.bad_request  # pylint: disable=no-member
            )
            err_msg = str(err).encode("utf-8")
            resp.raw = io.BytesIO(err_msg)
            resp.headers.update(
                {
                    "Content-Length": len(err_msg),
                    "Content-Type": "text/plain; charset=UTF-8",
                }
            )

        return resp

    @staticmethod
    def _set_file_headers(resp, path):
        """
        Set applicable HTTP headers for the given file.

        Args:
            resp:
                The requests.Response object.

            path:
                The path to the file as a pathlib.Path object.
        """
        stat = path.stat()
        mtime = datetime.datetime.fromtimestamp(stat.st_mtime, datetime.UTC)
        resp.headers.update(
            {
                "Content-Length": stat.st_size,
                "Last-Modified": mtime.strftime("%a, %d %b %Y %H:%M:%S GMT"),
            }
        )
        media_type, encoding = mimetypes.guess_type(path)
        if media_type:
            resp.headers["Content-Type"] = media_type
        if encoding:
            resp.headers["Content-Encoding"] = encoding

    def close(self):
        pass



#!/usr/bin/env python3
"""
IP functions.
"""

import logging
import socket

import psutil


LOGGER = logging.getLogger(__name__)


def local_ip_addresses():
    """
    A generator over local IP addresses. Note that some of these may refer to IP
    addresses of hosted containers, which may break assumptions about mapping
    file URIs to the host file system.
    """
    for addr_data in psutil.net_if_addrs().values():
        for item in addr_data:
            yield item.address


def is_localhost(host):
    """
    Check if a host appears to be localhost.

    Args:
        host:
            A hostname or IP address.
    """
    try:
        host = socket.gethostbyname(host)
    except socket.gaierror as err:
        LOGGER.debug('Failed to resolve host "%s": %s', host, err)
    return host in local_ip_addresses()



#!/usr/bin/env python3
"""
Test request_file_adapters.
"""

import contextlib
import datetime
import pathlib
import socket
import tempfile
import unittest

from requests_file_adapter import get_session

RFC_9110_FMT = "%a, %d %b %Y %H:%M:%S GMT"


@contextlib.contextmanager
def tmp_file():
    """
    Create a temporary file in a temporary directory.

    Returns:
        A 3-tuple with the file content, the file modification time and the
        path.
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir) / "tmp.txt"
        text = "placeholder text"
        tmp_path.write_text(text)
        mtime = datetime.datetime.fromtimestamp(
            tmp_path.stat().st_mtime, datetime.timezone.utc
        )
        yield text, mtime, tmp_path


class TestText(unittest.TestCase):
    """
    Test text loading.
    """

    def setUp(self):
        self._session = get_session()

    def test_file_content(self):
        """
        Retrieve file content.
        """
        with tmp_file() as (text, _, path):
            resp = self._session.get(path.as_uri())
            self.assertEqual(text, resp.content.decode())

    def test_head_request(self):
        """
        HEAD request sends no content.
        """
        with tmp_file() as (text, _, path):
            resp = self._session.head(path.as_uri())
            self.assertIsNone(resp.content)

    def test_headers(self):
        """
        Expected headers are set.
        """
        with tmp_file() as (content, mtime, path):
            for method in ("HEAD", "GET"):
                with self.subTest(method=method):
                    resp = getattr(self._session, method.lower())(path.as_uri())
                    headers = resp.headers
                    self.assertEqual(
                        headers["Content-Length"], len(content.encode("utf-8"))
                    )
                    self.assertEqual(
                        headers["Last-Modified"], mtime.strftime(RFC_9110_FMT)
                    )
                    self.assertEqual(headers["Content-Type"], "text/plain")

    def test_ok(self):
        """
        Successful response returns a 200 OK code.
        """
        with tmp_file() as (_, _, path):
            for method in ("HEAD", "GET"):
                with self.subTest(method=method):
                    resp = getattr(self._session, method.lower())(path.as_uri())
                    self.assertEqual(resp.status_code, 200)

    def test_missing(self):
        """
        Missing file returns a 404 Not Found error.
        """
        with tmp_file() as (_, _, path):
            path = path.parent / "missing.txt"
            for method in ("HEAD", "GET"):
                with self.subTest(method=method):
                    resp = getattr(self._session, method.lower())(path.as_uri())
                    self.assertEqual(resp.status_code, 404)

    def test_no_perm(self):
        """
        Permission error returns a 403 Forbidden error.
        """
        with tmp_file() as (_, _, path):
            path.chmod(0o000)
            for method in ("HEAD", "GET"):
                with self.subTest(method=method):
                    resp = getattr(self._session, method.lower())(path.as_uri())
                    self.assertEqual(resp.status_code, 403)

    def test_localhost(self):
        """
        Localhost is recognized.
        """
        hostname = socket.gethostname()
        with tmp_file() as (text, _, path):
            for host in (
                "localhost",
                "localhost.localdomain",
                "127.0.0.1",
                hostname,
                socket.gethostbyname(hostname),
            ):
                uri = f"file://{host}{path}"
                with self.subTest(host=host):
                    resp = self._session.get(uri)
                    self.assertEqual(text, resp.content.decode())

    def test_other_host(self):
        """
        Other hosts raise ValueErrors.
        """
        hostname = socket.gethostname() * 2
        with tmp_file() as (text, _, path):
            uri = f"file://{hostname}{path}"
            with self.assertRaises(ValueError):
                self._session.get(uri)

    def test_other_methods(self):
        """
        Unsupported HTTP methods raise ValueErrors.
        """
        with tmp_file() as (_, _, path):
            for method in ("POST", "PUT"):
                with self.subTest(method=method), self.assertRaises(ValueError):
                    getattr(self._session, method.lower())(path.as_uri())

    def test_rel_paths(self):
        """
        Relative paths raise value errors.
        """
        uri = "file://foo.txt"
        with self.assertRaises(ValueError):
            self._session.get(uri)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Utilities.
"""


import requests

from .adapter import FileAdapter


def get_session():
    """
    Get a requests.Session object with support for URI file schemes.
    """
    session = requests.Session()
    session.mount("file://", FileAdapter())
    return session
